﻿# Разработка скриптов на TypeScript 
Это небольшая инструкция, которая поможет скриптописателям повысить удобство и скорость
разработки скриптов для снегопата. Я предлагаю вам использовать для этих целей редактор
*Visual Studio Code* и язык *TypeScript*. Этот язык - развитие языка JavaScript,
к которому добавлена типизация переменных, классы и модули. Скрипт на языке TypeScript
транслируется его компилятором в обычный JavaScript.

Благодаря возможности подключить к редактору файлы с описанием SnegAPI и типов 1С,
при разработке скриптов для снегопата в редакторе отлично работает intellisense,
что делает работу быстрой и комфортной.  
Начать его использование для разработки помогут несколько простых шагов. 

1. Скачиваем и устанавливаем Visual Studio Code: https://code.visualstudio.com/
2. Скачиваем и устанавливаем TypeScript: http://www.typescriptlang.org/#Download  
    Я брал "TypeScript 1.6 in VS2015"
3. В папке снегопата, рядом с папками addins и core - создаем папку custom.
    Снегопат там ищет дополнительные пользовательские аддины.
4. Создаем папку для разработки, например "develop", можно тоже в папке снегопата,
    рядом с addins и custom.
5. В этой папке создаём подпапку std.
6. Копируем все *.ts файлы из core\scripts\src в папку develop\std.
7. В папке develop запускаем команду tsc --init
    Она создаст файл проекта tsconfig.json
8. В этом файле исправляем путь к каталогу сборки, параметр "outDir". Там надо прописать
    путь к каталогу custom. Если папка develop рядом с ним, просто укажите:
    "outDir": "../custom",
    И добавьте параметр "emitBOM": true
    Файл будет выглядеть примерно так
        {
            "compilerOptions": {
                "module": "commonjs",
                "target": "es3",
                "noImplicitAny": false,
                "outDir": "../custom",
                "rootDir": ".",
                "sourceMap": false,
                "emitBOM": true
            },
            "exclude": [
                "node_modules"
            ]
        }
9. Запускаем Visual Studio Code:
        code.exe "путь к папке develop"
    
10. Жмем Ctrl+Shift+P, набираем conf - выбираем "Task - configure task Runner"
    В открывшемся файле удаляем или комментируем первый пример (с HelloWorld.ts),
	и раскомментируем второй, который
        // A task runner that calls the Typescript compiler (tsc) and
        // compiles based on a tsconfig.json file that is present in
        // the root of the folder open in VSCode

11. В окне снегопата в меню "Разработка" выбираем "Сдампить SnegAPI в snegopat.d.ts"
    и "Сдампить типы 1С в v8.d.ts". Файлы создадуться в каталоге снегопата.
	Перемещаем их в каталог разработки (develop).
12. В-принципе, это всё. Теперь можно в этом каталоге создавать файлы с расширением .ts,
    и по нажатию Ctrl+Shift+B они будут компилится в js файлы в каталог custom.
	Вместе с ними также скомпилятся еще и std скрипты в папку custom\std, но они не помешают.

В начале своих скриптов пишем

        //engine: JScript
        //uname: уникальное имя
        //addin: global
        ... другие теги ...

Подключаем SnegAPI и типы 1С, файлы должны лежать в этом же каталоге:

        ...
        /// <reference path="./snegopat" />
        /// <reference path="./v8" />
        ...
Подключаем глобальные контексты 1С

        ...
        global.connectGlobals(SelfScript);
        ...
Подключаем нужные стандартные скрипты

        ...
        import * as stdlib from "./std/std";
        import * as stdcommands from "./std/commands";
        import * as hks from "./std/hotkeys";
        ...

и т.п., смотря какие скрипты нужны. В кавычках путь обязательно должен начинаться с точки, быть
относительным каталога, где лежит скрипт, расширение ts не указываем.
Также можно импортом подключать и другие свои скрипты, написанные на TypeScript.
В таких скриптах обязательно что-то должно быть export, смотрите примеры в стандартных скриптах.

По самому TypeScript.
Переменные типизируются или сами при объявлении присваиванием, если тип выводится из выражения,
либо можно через двоеточие указать явно. Самый общий тип - any.
Иногда TypeScript ругается на несовпадение типов - тогда приводите явно: <ИмяТипа>
Смотрите примеры, читайте интернеты. Успехов!

По отладке скриптов - смотрите тему https://snegopat.ru/forum/viewtopic.php?f=7&t=419&start=10#p7192
Вкратце - необходимо установить один или несколько из перечисленных скриптовых отладчиков,
в тексте скрипта в месте, где хочется подключить отладчик, написать debugger - и при выполнении
этой строки всплывёт отладчик. Дальше в нём уже можно выполнять пошагово, ставить точки останова,
просматривать переменные и т.д. и т.п.